package com.contextspace;

import static org.junit.Assert.*;

import java.io.Serializable;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.annotations.Index;
import javax.persistence.Basic;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.ManyToOne;
import javax.persistence.MappedSuperclass;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.Persistence;
import javax.persistence.TableGenerator;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.persistence.Transient;
import javax.persistence.TypedQuery;
import javax.persistence.Version;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.contextspace.addressTest.Attribute;
import com.contextspace.addressTest.AttributesHolder;
import com.contextspace.addressTest.IDHolder;


public class AddressTest {

    @Test
    public void test() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory("objectdb:$objectdb/db/test.tmp;drop");
        EntityManager em = emf.createEntityManager();
     
        
        em.getTransaction().begin();

        CDI cdi = new CDI("jbloggs","Joe", "Bloggs", new Date());
        em.persist(cdi);
     
        em.getTransaction().commit();
     
        TypedQuery<Long> query = em.createQuery("SELECT COUNT($1) FROM CDI $1 WHERE $1.loginId LIKE '%' ESCAPE '\\'", Long.class);
        Long cdiCount = query.getSingleResult();
        assertEquals(1, cdiCount.intValue());
        
        TypedQuery<CDI> query2 = em.createQuery("SELECT c FROM CDI c", CDI.class);
        List<CDI> results = query2.getResultList();
        assertEquals(1, results.size());
        
        System.out.println(printPrettyXml(results.get(0).toString(),4));
        
        InstantMessagingAddress instantMsg = new InstantMessagingAddress("home", "someaddress", "skype");
        cdi.addElectronicAddress(instantMsg);
        
        em.getTransaction().begin();
        em.merge(cdi);
        em.getTransaction().commit();
        
        TypedQuery<CDI> query3 = em.createQuery("SELECT c FROM CDI c", CDI.class);
        List<CDI> results2 = query3.getResultList();
        assertEquals(1, results2.size());
        System.out.println(printPrettyXml(results2.get(0).toString(),4));
        
        em.close();
        emf.close();
    }

    
    @Entity
    public static class CDI extends IDHolder {
        private static final long serialVersionUID = 1L;
       
        private Date creationDate = null;
        private String loginId = null;
        private String firstname = null;
        private String lastname = null;
        
        private ContactDetails contactDetails = new ContactDetails();
        
        private long id;
        
        /**
         * Default Constructor
         */
        public CDI() {
            
        }
        
        /**
         * Constructor
         * @param firstname
         * @param lastname
         */
        public CDI(String firstname, String lastname) {
            this.firstname = firstname;
            this.lastname = lastname;
            this.creationDate = new Date();
        }
        
        /**
         * Constructor
         * @param firstname
         * @param lastname
         * @param creationDate
         */
        public CDI(String loginId, String firstname, String lastname, Date creationDate) {
            this.loginId = loginId;
            this.firstname = firstname;
            this.lastname = lastname;
            this.creationDate = creationDate;
        }
            
        /**
         * @return the creationDate
         */
        @Temporal(TemporalType.TIMESTAMP)
        public Date getCreationDate() {
            return creationDate;
        }
        
        /**
         * @param creationDate the creationDate to set
         */
        public void setCreationDate(Date creationDate) {
            this.creationDate = creationDate;
        }
        
        /**
         * @return the object id
         */
        @Id
        @GeneratedValue(strategy=GenerationType.TABLE, generator="IdTableGen")
        public long getId() {
            return id;
        }

        /**
         * @param id the id to be set
         */
        public void setId(long id) {
            this.id = id;
        }

        /**
         * @return the firstname
         */
        @Index(unique="false")
        public String getFirstname() {
            return firstname;
        }

        /**
         * @param firstname the firstname to set
         */
        public void setFirstname(String firstname) {
            this.firstname = firstname;
        }

        /**
         * @return the lastname
         */
        @Index(unique="false")
        public String getLastname() {
            return lastname;
        }

        /**
         * @param lastname the lastname to set
         */
        public void setLastname(String lastname) {
            this.lastname = lastname;
        }

        /**
         * @return the loginId
         */
        @Index(unique="true")
        @Column(nullable=false,unique=true)
        public String getLoginId() {
            return loginId;
        }

        /**
         * @param loginId the loginId to set
         */
        public void setLoginId(String loginId) {
            this.loginId = loginId;
        }
        
        public void addElectronicAddress(ElectronicAddress address) {
            contactDetails.addElectronicAddress(address);
        }
        
        public void removeElectronicAddress(ElectronicAddress address) {
            contactDetails.removeElectronicAddress(address);
        }
        
        
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            
            buf.append("<cdi");
            if(getId() > 0) {
                buf.append(" id=\"").append(getId()).append("\"");
            }
            buf.append(">");
     
            if(firstname != null)
                buf.append("<firstname>").append(firstname).append("</firstname>");
            if(lastname != null)
                buf.append("<lastname>").append(lastname).append("</lastname>");
            if(loginId != null)
                buf.append("<loginid>").append(loginId).append("</loginid>");
            
            buf.append(contactDetails.toString());
            
            buf.append("</cdi>");
            return buf.toString();
        }

        /**
         * @return the contactDetails
         */
        @Embedded
        public ContactDetails getContactDetails() {
            return contactDetails;
        }

        /**
         * @param contactDetails the contactDetails to set
         */
        public void setContactDetails(ContactDetails contactDetails) {
            this.contactDetails = contactDetails;
        } 
        
        
    }
    
    @Embeddable
    public static class ContactDetails implements Serializable{
        private static final long serialVersionUID = -8395583852077269007L;

        private List<ElectronicAddress> electronicAddresses = new ArrayList<ElectronicAddress>();
        
        /**
         * Default Constructor
         */
        public ContactDetails() {
            
        }
        
        /**
         * @param emailAddresses
         * @param addresses
         * @param phoneNumbers
         */
        public ContactDetails(List<ElectronicAddress> addresses) {
            this.electronicAddresses = addresses;
        }
        
        /**
         * @param electronicAddress
         */
        public void addElectronicAddress(ElectronicAddress electronicAddress) {
            electronicAddresses.add(electronicAddress);
        }

        /**
         * @return the electronicAddresses
         */
        @OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER, orphanRemoval=true)
        public List<ElectronicAddress> getElectronicAddresses() {
            return electronicAddresses;
        }

        /**
         * @param electronicAddresses the electronicAddresses to set
         */
        public void setElectronicAddresses(List<ElectronicAddress> electronicAddresses) {
            this.electronicAddresses = electronicAddresses;
        }
        
        /**
         * @param electronicAddressToRemove
         */
        public void removeElectronicAddress(ElectronicAddress electronicAddressToRemove) {
            electronicAddresses.remove(electronicAddressToRemove);
        }
        
        /**
         * @param electronicAddressToReplace
         */
        public void replaceElectronicAddress(ElectronicAddress electronicAddressToReplace) {
            for(int i=0; i<electronicAddresses.size(); i++) {
                if(electronicAddresses.get(i).equals(electronicAddressToReplace)) {
                    electronicAddresses.set(i, electronicAddressToReplace);
                }
            }
        }
        
        /**
         * Clear Electronic Addresses
         */
        public void removeAllElectronicAddresses() {
            electronicAddresses.clear();
        }
        
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            if(electronicAddresses != null && !electronicAddresses.isEmpty()) {
                buf.append("<contactDetails>");
                if(electronicAddresses != null && !electronicAddresses.isEmpty()) {
                    for(Iterator<ElectronicAddress> i = electronicAddresses.iterator(); i.hasNext();) {
                        ElectronicAddress address = i.next();
                        buf.append(address.toString());
                    }
                }
                buf.append("</contactDetails>");     
            }
            return buf.toString();
        }
        
    
    }

    @MappedSuperclass
    public static abstract class AbstractContactDetail extends AttributesHolder {

      private static final long serialVersionUID = 1L;
      private boolean primary = false;
      private BasicMetaData basicMetaData = new BasicMetaData();
      
      protected String type;
      
      public AbstractContactDetail(){
          super();
          
          System.out.println("AbstractDetail Construct: " + (basicMetaData == null ? "null" : " not null"));
      }
      
      /**
       * @return the primary
       */
      @Column(name="isPrimary")
      @XmlAttribute(name="primary")
      public boolean getPrimary() {
          return primary;
      }
      
      /**
       * @param primary the primary to set
       */
      public void setPrimary(boolean primary) {
          this.primary = primary;
          
          setObjectMetaDataChange();
      }
      
      /**
       * @return the basicMetaData
       */
      @Embedded
      @XmlElement(name="basicMetaData")
      public BasicMetaData getBasicMetaData() {
          return basicMetaData;
      }

      /**
       * @param basicMetaData the basicMetaData to set
       */
      public void setBasicMetaData(BasicMetaData basicMetaData) {
          this.basicMetaData = basicMetaData;
      }
      
      /**
       * set object meta data change
       */
      protected void setObjectMetaDataChange() {
          // Set the metadata
          if(basicMetaData == null) {
              
              System.out.println("metadata is null: superclass: " + this.getClass().getSuperclass().getName());
          }
          
          basicMetaData.setLastModifiedBy("me");
          basicMetaData.setLastModifiedDate(new Date());
      }
      
      /**
       * @return the type
       */
      @Index
      public String getType() {
          return type;
      }

      /**
       * @param type the type to set
       */
      public void setType(String type) {
          this.type = type;
          
          setObjectMetaDataChange();
      }
      
      /* (non-Javadoc)
       * @see java.lang.Object#toString()
       */
      @Override
      public String toString() {
          StringBuffer buf = new StringBuffer();
          
          buf.append(" type=\"").append(type).append("\"");
          buf.append(" primary=\"").append(primary).append("\"");
          buf.append(">");
//          buf.append(super.toString());
          return buf.toString();
      }
      
      @Override
      public boolean equals(Object obj) {
          if(!(obj instanceof AbstractContactDetail)) {
              return false;
          }      
          AbstractContactDetail other = (AbstractContactDetail)obj;
          if(!super.equals(obj)){
              return false;
          }
          
          if(getPrimary() != other.getPrimary() ){
              return false;
          }
          if((getType()==null && other.getType()!=null) || (getType()!=null&&!getType().equals(other.getType())) ){
              return false;
          }
          
          return true;
      }
  }

  
  


  @Embeddable
  public static class BasicMetaData extends AbstractMetaData {
      
      /**
       * Basic Constructor
       */
      public BasicMetaData() {
          super();
      }
  }
  
  @MappedSuperclass
  public static abstract class AbstractMetaData implements Serializable{

      private Date creationDate = null;
      private Date lastModifiedDate = null;
      private String lastModifiedBy = null;

      /**
       * Default Constructor
       * Set creation date
       */
      public AbstractMetaData() {
          super();
          this.creationDate = new Date();
      }
      
      /**
       * @return the creationDate
       */
      @Temporal(TemporalType.TIMESTAMP)
      @XmlAttribute(name = "creationDate")
      public Date getCreationDate() {
          return creationDate;
      }
      
      
      /**
       * @param creationDate the creationDate to set
       */
      public void setCreationDate(Date creationDate) {
          this.creationDate = creationDate;
      }
      
      /**
       * @return the lastModifiedDate
       */
      @Temporal(TemporalType.TIMESTAMP)
      @XmlAttribute(name = "lastModifiedDate")
      public Date getLastModifiedDate() {
          return lastModifiedDate;
      }
      /**
       * @param lastModifiedDate the lastModifiedDate to set
       */
      public void setLastModifiedDate(Date lastModifiedDate) {
          this.lastModifiedDate = lastModifiedDate;
      }
      
      /**
       * @return the lastModifiedBy
       */
      @XmlAttribute(name = "lastModifiedBy")
      public String getLastModifiedBy() {
          return lastModifiedBy;
      }
      /**
       * @param lastModifiedBy the lastModifiedBy to set
       */
      public void setLastModifiedBy(String lastModifiedBy) {
          this.lastModifiedBy = lastModifiedBy;
      }
      
  }

  

  @MappedSuperclass
  public static abstract class ElectronicAddress extends AbstractContactDetail {

      private static final long serialVersionUID = 1L;
      
      private String address;
      private String domain;  //top level domain for the address provider, e.g. facebook.com

      /**
       * Default Constructor
       */
      protected ElectronicAddress() {
          super();
      }
      
      protected ElectronicAddress( String address) {
          super();
          this.address = address;
      }

      /**
       * Constructor taking type and its value
       * @param type
       * @param value
       */
      protected ElectronicAddress(String type, String address) {
          super();
          this.type = type;
          this.address = address;
      }

      /**
       * Constructor taking type and its value
       * @param type
       * @param value
       */
      protected ElectronicAddress(String type, String address, String domain) {
          super();
          this.type = type;
          this.address = address;
          this.domain = domain==null?null:domain.toLowerCase();
      }

      /**
       * @return the value
       */
      @Index(unique="false")
      public String getAddress() {
          return address;
      }

      /**
       * @param value the value to set
       */
      public void setAddress(String address) {
          this.address = address;
          setObjectMetaDataChange();
      }
      
      public String getDomain() {
          return domain;
      }

      public void setDomain(String domain) {
          this.domain = domain==null?null:domain.toLowerCase();
          setObjectMetaDataChange();
      }

      @Override
      public String toString() {
          StringBuffer buf = new StringBuffer();
          buf.append(super.toString()); 
          buf.append("<address>").append(getAddress()).append("</address>");
          buf.append("<domain>").append(domain).append("</domain>");               
          return buf.toString();
      }
      
      /* (non-Javadoc)
       * @see java.lang.Object#equals(java.lang.Object)
       */
      @Override
      public boolean equals(Object obj) {
          if(!(obj instanceof ElectronicAddress)) {
              return false;
          }
          if(!super.equals(obj)){
              return false;
          }
          ElectronicAddress otherElectronicAddress = (ElectronicAddress)obj;

          if( (getAddress()==null && otherElectronicAddress.getAddress() !=null) ||
                  (getAddress()!=null && !getAddress().equalsIgnoreCase(otherElectronicAddress.getAddress())) ) {
              return false;
          }
          if( (getDomain()==null && otherElectronicAddress.getDomain() !=null) ||
                  (getDomain()!=null && !getDomain().equalsIgnoreCase(otherElectronicAddress.getDomain())) ) {
              return false;
          }
          return true;
      } 
      
  }
    
  
  @Entity
  public static class InstantMessagingAddress extends ElectronicAddress {
      public static enum IM_PROVIDERS {aim, gtalk, icq, xmpp, msn, skype, qq, yahoo};
      
      public InstantMessagingAddress() {
          super();
      }

      public InstantMessagingAddress( String address) {
          super(address);
      }

      public InstantMessagingAddress(String type, String address) {
          super(type, address);
      }

      public InstantMessagingAddress(String type, String address, String domain) {
          super(type, address, domain);       
          validateDomain(domain);
      }

      @Override
      public void setDomain(String domain) {
          validateDomain(domain);
          super.setDomain(domain);
      }

      
      @Override
          public String toString() {
              StringBuffer buf = new StringBuffer();
              buf.append("<instantMessagingAddress");
              buf.append(super.toString());        
              buf.append("</instantMessagingAddress>");
              return buf.toString();
          }

          @Override
          public boolean equals(Object obj) {
              if(!(obj instanceof InstantMessagingAddress)) {
                  return false;
              }
              return super.equals(obj);
          }
          
          public static IM_PROVIDERS validateDomain(String domain){
              System.out.println("Domamin: " + domain);
              return IM_PROVIDERS.valueOf(domain);            
          }

  }

  @MappedSuperclass
  @TableGenerator(name="IdTableGen", initialValue=1, allocationSize=50)
  public static abstract class IDHolder implements Serializable{
      private static final long serialVersionUID = 1L;
      

      private long version;

      private long id;

      /**
       * @return the object id
       */
      @Id
      @GeneratedValue(strategy=GenerationType.TABLE, generator="IdTableGen")
      public long getId() {
          return id;
      }

      /**
       * @param id the id to be set
       */
      public void setId(long id) {
          this.id = id;
      }

      @Version
      public long getVersion() {
          return version;
      }
      
      
  }   
  
  
  @Entity
  @TableGenerator(name="IdAttrGen", initialValue=1, allocationSize=50)
  public static class Attribute implements Serializable {

      private static final long serialVersionUID = 1L;
      static Logger log = LoggerFactory.getLogger(Attribute.class.getName());
      private long id=0;
      
      private AttributesHolder owningObject;
      private String name;
      private String value;
      private String type;
      private String displayName;
      private String language = "en";
      private boolean primary = false;


      
      /**
       * Constructor 
       * @param name
       * @param values
       */
      public Attribute(String name, Object value, String language) {
          this.name = name;
          this.language = language;
          this.type = value.getClass().getName();
          this.value = convertValueToString(value);
      }
      
      /**
       * Constructor 
       * @param name
       * @param values
       */
      public Attribute(String name, Object value) {
          this.name = name;
          this.type = value.getClass().getName();
          this.value = convertValueToString(value);
      }
      
      /**
       * @return the object id
       */
      @Id
      @GeneratedValue(strategy=GenerationType.TABLE, generator="IdAttrGen")
      public long getId() {
          return id;
      }

      /**
       * @param id the id to be set
       */
      public void setId(long id) {
          this.id = id;
      }
      
      /**
       * Constructor 
       * @param nam
       */
      public Attribute(String name) {
          this.name = name;
      }
      
      /**
       * Default constructor
       */
      public Attribute() {
      }
      
      /**
       * @return the name
       */
      @Index
      @Basic
      @XmlAttribute(name = "name")
      public String getName() {
          return name;
      }
      /**
       * @param name the name to set
       */
      public void setName(String name) {
          this.name = name;
      }

      /**
       * @return the values
       */
      @XmlTransient
      @Transient
      public Object getValue() {
          if(type.equals(String.class.getName())) {
              return value;
          } else if(type.equals(Boolean.class.getName())) {
              return new Boolean(value);
          } else if(type.equals(Integer.class.getName())) {
              return new Integer(value);
          } else if(type.equals(Long.class.getName())) {
              return new Long(value);
          } else {
              try {
                  return Class.forName(type).getConstructor(String.class).newInstance(value);
              } catch (Exception e) {
                  log.error("Failed to recreate object in attribute: " + e.getMessage(), e);
                  return null;
              }
          }
      }


      public static String convertValueToString(Object value){
          return value.toString();
      }
      /**
       * @param values the values to set
       */
      public void setValue(Object value) {
          if(value != null) {
              this.type = value.getClass().getName();
              this.value = convertValueToString(value);
          }
      }
      
      @Index
      @Basic
      public String getValueAsString() {
          return value;
      }
      
      public void setValueAsString(String value) {
          this.value = value;
      }

      /**
       * @return the displayName
       */
      @XmlAttribute(name = "displayname")
      public String getDisplayName() {
          return displayName;
      }


      /**
       * @param displayName the displayName to set
       */
      public void setDisplayName(String displayName) {
          this.displayName = displayName;
      }
      
      @Basic
      @XmlAttribute(name = "type")
      public String getType() {
          return type;
      }

      public void setType(String type) {
          this.type = type;
      }

      /**
       * @return the owningObjectId
       */
      @ManyToOne
      public AttributesHolder getOwningObject() {
          return owningObject;
      }

      /**
       * @param owningObjectId the owningObjectId to set
       */
      public void setOwningObject(AttributesHolder owningObject) {
          this.owningObject = owningObject;
      }

      /**
       * @return the language
       */
      public String getLanguage() {
          return language;
      }

      /**
       * @param language the language to set
       */
      public void setLanguage(String language) {
          this.language = language;
      }

      public boolean isPrimary() {
          return primary;
      }

      public void setPrimary(boolean primary) {
          this.primary = primary;
      }

      /* (non-Javadoc)
       * @see java.lang.Object#toString()
       */
      @Override
      public String toString() {
          StringBuffer buf = new StringBuffer();
          buf.append("<attribute name=\"").append(name).append("\" language=\"").append(language);
          buf.append("\" type=\"").append(type);
          buf.append("\" value=\"").append(value).append("\"");
          if(getId() != 0) {
              buf.append(" id=\"").append(getId()).append("\"");
          }
          if(displayName != null) {
              buf.append(" displayName=\"").append(displayName).append("\"");
          }
          if(owningObject != null) {
              buf.append(" owningid=\"").append(owningObject.getId()).append("\"");
          }
          buf.append("/>");      
          return buf.toString();
      }
          
      @Override
      public boolean equals(Object obj) {
          //log.debug("Attribute Equals Called: ");
          if (obj == null || (getClass() != obj.getClass())) {
              return false;
          }
          if (this == obj) {
              return true;
          }
          
          Attribute other = (Attribute) obj;
          
          //log.debug("Comparing: " + name + " with " + other.getName());
          //log.debug("Comparing: " + value + " with " + other.getValue());
          //log.debug("Comparing: " + language + " with " + other.getLanguage());
          //log.debug("Comparing: " + getId() + " with " + other.getId());
          
          if((getId() != 0 && getId() == other.getId())) {
              return true;
          }

          if ( (value == null && other.getValue() != null) || (name == null && other.getName() != null) ||
                  (language == null && other.getLanguage() != null) ||  (type == null && other.getType() != null) ) {
              //log.debug("Not Equals");
              return false;                
          } else if ((name==null || name.equals(other.getName()) && (value==null || value.equals(other.getValue()))
                  && (language==null || language.equals(other.getLanguage())) && (type==null || type.equals(other.getType())) 
                   && primary==other.isPrimary()))  {
              //log.debug("Equals");
              return true;
          }
          //log.debug("Not Equals");
          return false;
      }
  }
  
  @Entity
  public static class AttributesHolder extends IDHolder {
      private static final long serialVersionUID = 1L;

      static Logger log = LoggerFactory.getLogger(AttributesHolder.class.getName());
      
      private Map<String,List<Attribute>> attributeMap = new HashMap<String, List<Attribute>>();
      private List<Attribute> attributeList = new ArrayList<Attribute>();
      
      public AttributesHolder(){
      }
      
      /**
       * Helper method to populate the quick lookup map
       * @param attr
       */
      private void populateMap(Attribute attr) {
          String key = attr.getName();
          if(attributeMap.containsKey(key)) {
              attributeMap.get(key).add(attr);
          }
          else {
              List<Attribute> newList = new ArrayList<Attribute>();
              newList.add(attr);
              attributeMap.put(key, newList);
          }
      }
      
      /**
       * Check the the quick lookup map is populated
       */
      private void checkMap() {
          if(attributeMap.isEmpty()) {
              for(int i=0; i < attributeList.size();i++) {
                  populateMap(attributeList.get(i));
              }
          }
      }
      
      public void printMap() {
          if(log.isDebugEnabled()) {
              for(Iterator<String> i = attributeMap.keySet().iterator(); i.hasNext();) {
                  String key = i.next();
                  List<Attribute> attrs = attributeMap.get(key);
                  for(Attribute attr : attrs) {
                      log.debug("Key: " + key + " Attr: " + attr);
                  }
              }
          }
      }
      
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#getAttribute(java.lang.String)
       */
      public List<Attribute> getAttributes(String name) {
          checkMap();
          
          String[] nameLangPair = name.split(",lang=");
          if(nameLangPair.length > 1) {
              try {
                  return getAttributes(nameLangPair[0], nameLangPair[1]);
              } catch (Exception e) {
                  log.error("Exception: ", e);
              }
          }
          return attributeMap.get(name);
      }
      
      public Set<String> findAttributeNames() {
          checkMap();
          return attributeMap.keySet();
      }
      
      public List<Attribute> getAttributesForLanguage(String language) throws Exception {
          List<Attribute> returnList = new ArrayList<Attribute>();
          for(Attribute attr : attributeList) {
              if(attr.getLanguage().equals(language)) {
                  returnList.add(attr);
              }
          }
          return returnList;
      }
      
      public List<Attribute> getAttributes(String name, String language) throws Exception {
          return getAttributes(name, language, true);     
      }
      
      public List<Attribute> getAbsoluteAttributes(String name, String language) throws Exception {
          return getAttributes(name, language, false);     
      }
      
      
      private List<Attribute> getAttributes(String name, String language, boolean returnEnglishIfNoneFound) throws Exception {
          
          checkMap();
          
          List<Attribute> attributes = attributeMap.get(name);
          
          // just return null if the attribute doesn't exist
          if(attributes == null || attributes.isEmpty()) {
              return null;
          }
          
          List<Attribute> rightLanguageAttributes = new ArrayList<Attribute>();
          
          // return the values in the specified language
          for(Attribute attr:attributes) {
              if(attr.getLanguage().equals(language)) {
                  rightLanguageAttributes.add(attr);
              }
          }

          // If the list is empty, the language for this attribute doesn't exist. Try to return 'en' by default
          if(rightLanguageAttributes.isEmpty() && returnEnglishIfNoneFound) {
              for(Attribute attr:attributes) {
                  if(attr.getLanguage().equals("en")) {
                      rightLanguageAttributes.add(attr);
                  }
              }
          }
          
          // if the list is still empty, return the original list
          if(rightLanguageAttributes.isEmpty() && returnEnglishIfNoneFound) {
              return attributes;
          }
          
          if(rightLanguageAttributes == null || rightLanguageAttributes.isEmpty()) {
              return null;
          }
          return rightLanguageAttributes;
      }
      
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#findAttribute(java.lang.String, java.lang.Object)
       */
      public Attribute findAttribute(String name, Object value) {
          checkMap();
          List<Attribute> attrList = attributeMap.get(name);
          for(int i=0; i<attrList.size(); i++) {
              Attribute attr = attrList.get(i);
              if(attr.getValue().equals(value)) {
                  return attr;
              }
          }
          return null;
      }
     
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#addAttribute(com.contextspace.datamodel.attribute.Attribute)
       */
      public void addAttribute(Attribute attribute) {
          attribute.setOwningObject(this);
          attributeList.add(attribute);
          populateMap(attribute);
      }
      
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#removeAttribute(java.lang.String)
       */
      public void removeAttribute(String name) {
          attributeMap.remove(name);
          
          List<Attribute> removalList = new ArrayList<Attribute>();
          for(int i = 0;i < attributeList.size(); i++) {
              if(attributeList.get(i).getName().equals(name)) {
                  removalList.add(attributeList.get(i));
              }
          }   
          log.debug("Found " + removalList.size() + " attributes to remove - current size: " + attributeList.size());
          attributeList.removeAll(removalList);
          log.debug("After remove: " + attributeList.size());
      }
          
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#removeAttribute(com.contextspace.datamodel.attribute.Attribute)
       */
      public void removeAttribute(Attribute attributeToRemove) {
          String name = attributeToRemove.getName();
          String value = attributeToRemove.getValueAsString();
          String lang = attributeToRemove.getLanguage();
          
          List<Attribute> remainingList = new ArrayList<Attribute>();
          List<Attribute> removalList = new ArrayList<Attribute>();
          
          for(int i = 0;i < attributeList.size(); i++) {
              Attribute attr = attributeList.get(i);
              
              //remember attribute list contains everything, so the remainingList should ONLY contain
              //things that have the same name
              if(attr.getName().equals(name) && attr.getLanguage().equals(lang) && attr.getValueAsString().equals(value)) {
                  removalList.add(attr);
              }
              else if(attr.getName().equals(name)){
                  remainingList.add(attr);    
              }
          }
          log.debug("Found " + removalList.size() + " attributes to remove - current size: " + attributeList.size());
          attributeList.removeAll(removalList);
          log.debug("After remove: " + attributeList.size());
          
          if(remainingList.isEmpty()) {
              attributeMap.remove(name);
          }
          else {
              attributeMap.put(name, remainingList);
          }
      }
      
      public void replaceAttribute(Attribute attributeToReplace) throws Exception {
          attributeToReplace.setOwningObject(this);
          
          boolean found = false;
          for(int i = 0;i < attributeList.size(); i++) {
              Attribute attr = attributeList.get(i);
              
              if((attr.getId() != 0 && attr.getId() == attributeToReplace.getId()) || 
                      (attr.getName().equals(attributeToReplace.getName()) && attr.getLanguage().equals(attributeToReplace.getLanguage()))) {
   
              //if(attr.equals(attributeToReplace)) {
                  attributeList.set(i, attributeToReplace);
                  found = true;
              }
          }
          
          //TODO may need to update the map with the fresh attribute
          
          if(!found) {
              throw new Exception("Failed to find attribute with id=" + attributeToReplace.getId() + " during replaceAttribute operation");
          }
      }
      

      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#getAttributeList()
       */
      @OneToMany(targetEntity = Attribute.class, mappedBy = "owningObject", cascade = CascadeType.ALL, fetch = FetchType.EAGER, orphanRemoval=true)
      @XmlElementWrapper(name="attributes")
      @XmlElement(name = "attribute")
      public List<Attribute> getAttributeList() {
          return attributeList;
      }
      
      /* (non-Javadoc)
       * @see com.contextspace.datamodel.attribute.IAttributesHolder#setAttributeList(java.util.List)
       */
      public void setAttributeList(List<Attribute> attrList) {
          attributeMap.clear();
          
          attributeList = attrList;
          
          //if this has been set by spring, then its likely that the link between this (the owningobject) and the attribute
          //hasn't been set. Spin the attributes setting it
          for(Attribute attr : attributeList) {
              attr.setOwningObject(this);
          }
      }
      
      /* (non-Javadoc)
       * @see java.lang.Object#toString()
       */
      @Override
      public String toString() {
          StringBuffer buf = new StringBuffer();
          
          if(attributeList!=null&&!attributeList.isEmpty()) {
              buf.append("<attributes>");
              for(Attribute attr: attributeList) {
                  buf.append(attr.toString());
              }
              buf.append("</attributes>");
          }
          return buf.toString();
      }


      @Override
      public boolean equals(Object obj) { 
          if(!(obj instanceof AttributesHolder)) {
              return false;
          }
          
          AttributesHolder other = (AttributesHolder)obj;
          
          if(getId()!=0 && other.getId()!=0) {
              return (getId()==other.getId());
          }
          
          checkMap();
          other.checkMap();
          for(String name:attributeMap.keySet()){
              List<Attribute> list = attributeMap.get(name);
              List<Attribute> otherList = other.attributeMap.get(name);
              if(otherList==null || list.size()!=otherList.size()){
                  return false;
              }
              for(Attribute attr:list){
                  boolean found = false;
                  for(Attribute otherattr:otherList){
                      if(attr.getValueAsString().equals(otherattr.getValueAsString())){
                          found=true;
                          break;
                      }
                  }
                  if(!found){
                      return false;
                  }
              }
          }
          return true;       
      }
  }

  
    
    /**
     * Print pretty xml
     * @param input
     * @param indent
     * @return
     * @throws TransformerException
     */
    public static String printPrettyXml(String input, int indent) {
        try {
            Source xmlInput = new StreamSource(new StringReader(input));
            StringWriter stringWriter = new StringWriter();
            StreamResult xmlOutput = new StreamResult(stringWriter);
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();
            
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", String.valueOf(indent));
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.transform(xmlInput, xmlOutput);
            return xmlOutput.getWriter().toString();
        }
        catch(TransformerException e) {
            return "Exception thrown during transform of xml: " + e.getMessage();
        }
    }
}
